/* 
   RS232 temperature sensor board
   ------------------------------

   clock is defined on gcc command line using 'DF_CPU=16000000UL'


                      ATtiny85
               
                      +--- ---+
                 PB5 -|1 |_| 8|- VCC 
                 PB3 -|2     7|- PB2  ------------- -->   RX (PC)          
                 PB4 -|3     6|- PB1  ------------- <--   TX (PC)           
                 GND -|4     5|- PB0  
                      +-------+

   Code by: M. Post
   Last modified: 11 June 2011

   something on progmem:
   http://www.avrfreaks.net/index.php?name=PNphpBB2&file=viewtopic&t=38003



 */


#include <avr/io.h>
#include <util/delay.h>
#include <string.h>
#include <math.h>
#include <avr/pgmspace.h>

// RS232 pin definitions
#define Sout PB1    // Define Serial-Out pin 

#define HIGH   1
#define LOW    0
#define INPUT  1 
#define OUTPUT 0
#define MARK   0 
#define SPACE  1

#define THERM_PORT   PORTB
#define THERM_DDR    DDRB
#define THERM_PIN    PINB
#define THERM_DQ     PB0        // Define DS1820 DQ pin

#define THERM_INPUT_MODE()      THERM_DDR&=~(1<<THERM_DQ)
#define THERM_OUTPUT_MODE()     THERM_DDR|=(1<<THERM_DQ)
#define THERM_LOW()             THERM_PORT&=~(1<<THERM_DQ)
#define THERM_HIGH()            THERM_PORT|=(1<<THERM_DQ)

#define THERM_CMD_CONVERTTEMP    0x44
#define THERM_CMD_RSCRATCHPAD    0xbe
#define THERM_CMD_WSCRATCHPAD    0x4e
#define THERM_CMD_CPYSCRATCHPAD  0x48
#define THERM_CMD_RECEEPROM      0xb8
#define THERM_CMD_RPWRSUPPLY     0xb4
#define THERM_CMD_SEARCHROM      0xf0
#define THERM_CMD_READROM        0x33
#define THERM_CMD_MATCHROM       0x55
#define THERM_CMD_SKIPROM        0xcc
#define THERM_CMD_ALARMSEARCH    0xec

#define THERM_DECIMAL_STEPS_12BIT   625    // 0.0625

#define PRESENCE_ERR 0xFF
#define DATA_ERR     0xFF
#define LAST_DEVICE  0x00
#define SEARCH_FIRST 0xFF

#define CLEAR(s) memset(&(s), 0, sizeof(s))     // to clear an array


int      got_ack;
uint8_t  romcodes[20][9];  // store romcode of 20 sensors max
uint8_t  scratchpad[9];    // stores a scratchpad

// write constants to EEPROM instead of RAM to save space
prog_char msg00[] PROGMEM = {"Short Circuit!"};
prog_char msg01[] PROGMEM = {"No sensors found"};
prog_char msg02[] PROGMEM = {"Bus Error!"};
prog_char msg03[] PROGMEM = {"Device "};
prog_char msg04[] PROGMEM = {"Devices: "};
prog_char msg05[] PROGMEM = {"Temp: "};
prog_char msg06[] PROGMEM = {""};

PROGMEM const char *msg[] = 	  
{
  msg00,
  msg01,
  msg02,
  msg03,
  msg04,
  msg05,
  msg06,
};

char buffer[16];



void pinMode(int pin,int state)
{ 
  if (state == OUTPUT)
  {
    (DDRB |= (1 << pin)); 
  } else { 
    (DDRB &= ~(1 << pin)); 
  }
} 

void digitalWrite(int pin, int state)
{
  if (state == HIGH)
  { 
    (PORTB |= (1 << pin)); 
  } else {
    (PORTB &= ~(1 << pin));
  }
} 


int digitalRead(int pin)
{ 
  if (bit_is_set(PINB,pin))
  {
    return 1; 
  } else { 
    return 0; 
  }
} 


// --  begin RS232 functions --

void initialise_Serial(long int baud)
{

  pinMode(Sout,LOW);

  // work out mark and space lengths based on clock speed

  /*
        bps     bit length (us)
   --------------------------
        2400       416.7
        4800       208.3
        9600       104.2
       19200        52.1
       38400        26.0
       57600        17.4
      115200         8.7

  */
 
 
}

void Write_Serial_Data(unsigned short D)
{

  // start bit is high, stop bit is low
  // character bits: 0=high, 1=low
  // mark=0 (LOW, 0 Volts)
  // space=1 (HIGH, 5 Volts)

  // #define CHECK_BIT(var,pos) ((var) & (1<<(pos)))   // not used anymore
  // if (CHECK_BIT(D,t))

  digitalWrite(Sout,MARK);    // start bit
  _delay_us(104);              

  int t;
  for (t=0 ; t <=7 ; t++)
  {
   if ((D) & (1<<(t)))
   {
      digitalWrite(Sout,SPACE);
    } else {
      digitalWrite(Sout,MARK);
    }  
    _delay_us(104);        
  }

  digitalWrite(Sout,SPACE); // stop bit 
  _delay_us(104);           

}

void Write_Serial_Text(char *StrData)
{

  int p = 0;
  int q = strlen(StrData);
  int temp = 0;

  for (p = 0; p < q; p++)
  {
    temp = StrData[p];
    Write_Serial_Data(temp);
  }
}

void Write_Serial_Number(int number)
{
  int digits;
  
  if (number == 0)
  {
    digits = 0;
    int array[0];
    array[0]=0;
    Write_Serial_Data(array[0]+48);
  
  } else {

    // determine the number of digits
    digits = (int) (log(number)/log(10))+1;
    
    // split up the number's digits into an array
    int i = digits -1;
    int array[digits];
    while (number > 0)
    { 
      array[i--] = number % 10;
      number /= 10;
    }

    // send array over i2c
    for (i =0; i <= digits-1; i++)
    {
      Write_Serial_Data(array[i]+48);
    }

  }
}

void Write_Serial_Hex(uint8_t x)
{

  // sends one byte in hex (without 0x prefix) 

  char nibble;

  /* extract high nibble */
  nibble = x & 0xf0;        // (resets lower nibble to all zero's)
  nibble = x >> 4;          // (moves high nibble to lower nibble)

  // output the high nibble 
  if(nibble < 10) 
  {
    Write_Serial_Data(48 + nibble);
  } else {
    Write_Serial_Data(65 + nibble - 10);
  }

  /* do the same on the low nibble */
  nibble = x & 0x0f;

  if(nibble < 10) 
  {
    Write_Serial_Data(48 + nibble);
  } else {
    Write_Serial_Data(65 + nibble - 10);
  }

}

// --  end RS232 functions --




// -- Begin OneWire functions --

inline __attribute__((gnu_inline)) void therm_delay(uint16_t delay)
{
  while(delay--) asm volatile("nop");
}

uint8_t owReset(int Pin)
{
  uint8_t i;

  digitalWrite(Pin, HIGH);
  pinMode(Pin, INPUT);

  digitalWrite(Pin, LOW);  
  pinMode(Pin, OUTPUT);    // bring low for 480us

  _delay_us(480);
 
  // Release line and wait for 60us
  digitalWrite(Pin, HIGH);
  pinMode(Pin, INPUT);
  _delay_us(60);

  // store line value and wait for completion of 480us period
  i=digitalRead(Pin);

  _delay_us(420);
  
  // return measured line value
  return i;

}

uint8_t therm_reset()
{
  uint8_t i;

  // pull line low and wait for 480us
  THERM_LOW();
  THERM_OUTPUT_MODE();
  therm_delay(480);
 
  // Release line and wait for 60us
  THERM_INPUT_MODE();
  therm_delay(60);

  // store line value and wait for completion of 480us period
  i=(THERM_PIN & (1<<THERM_DQ));
  therm_delay(420);
  
  // return measured line value
  return i;

}



void therm_write_bit(uint8_t bit)
{
  // pull line low for 1us
  THERM_LOW();
  THERM_OUTPUT_MODE();
  therm_delay(1);
  
  // if we want to write 1, release the line (if not will keep low)
  if (bit) THERM_INPUT_MODE();  

  // wait for 60us and release the line
  therm_delay(60);
  THERM_INPUT_MODE();

}

uint8_t therm_read_bit(void)
{
  uint8_t bit=0;

  // pull line low for 1us
  THERM_LOW();
  THERM_OUTPUT_MODE();
  therm_delay(1);

  // release line and wait for 14us
  THERM_INPUT_MODE();
  therm_delay(14);

  // read line value
  if (THERM_PIN&(1<<THERM_DQ)) bit=1;

  // wait for 45us to end and return read value
  therm_delay(45);

  return bit;

}

uint8_t therm_read_byte(void)
{

  uint8_t i=8, n=0;
 
  while(i--)
  {
    // shift one position right and store read value
    n>>=1;
    n|=(therm_read_bit()<<7);
  }

  return n;
}

void therm_write_byte(uint8_t byte)
{
  uint8_t i=8;

  while(i--)
  {
    // write actual bit and shift one position right to make the next bit ready
    therm_write_bit(byte&1);
    byte>>=1;
  }

}



unsigned char therm_rom_search(unsigned char diff, unsigned char *id)
{
  // scans a one-wire bus for all attached DS18x20 sensors

  unsigned char i, j, next_diff;
  unsigned char b;

  if( therm_reset() )
    return PRESENCE_ERR;                       // error, no device found
  therm_write_byte( THERM_CMD_SEARCHROM );     // ROM search command
  next_diff = LAST_DEVICE;                     // unchanged on last device
  i = 8 * 8;                        // 8 bytes
  do
  {
    j = 8;                          // 8 bits
    do
    {
      b = therm_read_bit();      // read bit

      // read complement bit
      if( therm_read_bit() ) 
      {         
        if( b )                     // 11
          return DATA_ERR;          // data error
      } else {
        if( !b )
        {
          // 00 = 2 devices
          if( diff > i || ((*id & 1) && diff != i) )
          {
            b = 1;                  // now 1
            next_diff = i;          // next pass 0
          }
        }
      }
      therm_write_bit( b );         // write bit
      *id >>= 1;
      if( b )                       // store bit
        *id |= 0x80;
      i--;
    } while( --j );

    id++;                           // next byte
  } while( i );

  return next_diff;                 // to continue search

} 




void start_meas( void )
{
  if( PINB & 1<< THERM_DQ )
  {
    therm_write_byte(THERM_CMD_CONVERTTEMP);    // read scratchpad

    PORTB |= 1<< THERM_DQ;
    DDRB |= 1<< THERM_DQ;			// parasite power on

  } else {

    strcpy_P(buffer, (char*)pgm_read_word(&(msg[00]))); 
    Write_Serial_Text(buffer);  // 'Short circuit!'
    _delay_ms(1000);

  }
}


int scanbus(void)
{

  // outputs the number of found devices and stores all roms in an array
  unsigned char id[8], diff;
  unsigned char i,x;
  unsigned char device_counter=0;

  // clear array

  for (x=1; x <= 20; x++)
  {
    romcodes[x][0] = 0;
  }


  for( diff = SEARCH_FIRST; diff != LAST_DEVICE; )
  {
    diff = therm_rom_search( diff, id );

    if( diff == PRESENCE_ERR )
    {
      strcpy_P(buffer, (char*)pgm_read_word(&(msg[01]))); 
      Write_Serial_Text(buffer);  // 'No sensors found'
      _delay_ms(1000);
      break;
    }
    if( diff == DATA_ERR )
    {
      strcpy_P(buffer, (char*)pgm_read_word(&(msg[02]))); 
      Write_Serial_Text(buffer);  // 'Bus Error'
      _delay_ms(1000);
      break;
    }

    if( id[0] == 0x28 || id[0] == 0x10 )
    {
      // valid temperature sensor found

      // store romcode in array
      for( i = 0; i < 8; i++ )
      {

        romcodes[device_counter][i]=id[i];
        Write_Serial_Hex(id[i]);

      }

      device_counter=device_counter+1;

    }
  }

  return device_counter;
}

void therm_romtemp(uint8_t device)
{
  // address a specific sensor and retrieve its temperature

  uint8_t i;

  therm_reset();
  therm_write_byte(THERM_CMD_MATCHROM);
  for (i=0 ; i < 8; i++)
  {
    therm_write_byte(romcodes[device][i]);
  }
  therm_write_byte(THERM_CMD_CONVERTTEMP);

  // wait until conversion is complete
  while(!therm_read_bit());



  therm_reset();
  therm_write_byte(THERM_CMD_MATCHROM);
  for (i=0 ; i < 8; i++)
  {
    therm_write_byte(romcodes[device][i]);
  }
  therm_write_byte(THERM_CMD_RSCRATCHPAD);

  // store the replies in the 'scratchpad' variable
  for (i=0 ; i < 2; i++)
  {
    scratchpad[i]=therm_read_byte();
  }
  therm_reset();

}


void showdevs()
{
  // outputs all device romcodes stored in the array

  unsigned char device=0;
  unsigned char rombyte=0;
  //uint8_t digit;
  uint16_t decimal;
 
  while (1)
  {

    if (romcodes[device][0] != 0)
    {

      // valid romcode found in array 

      // target the sensor and read its temperature
      therm_romtemp(device);   // this fills the 'scratchpad' variable

      strcpy_P(buffer, (char*)pgm_read_word(&(msg[03]))); 
      Write_Serial_Text(buffer);  // 'Device '
      Write_Serial_Number(device+1);

      for (rombyte=0; rombyte < 8; rombyte++)
      {    
        Write_Serial_Hex(romcodes[device][rombyte]);            // display the romcode as hex
      } 


      // store temperature integer digits and decimal digits

      //digit=scratchpad[0]>>4;
      //digit|=(scratchpad[1]&0x7)<<4;
      // store decimal digits
      //decimal=scratchpad[0]&0xf;
      //decimal*=THERM_DECIMAL_STEPS_12BIT;


      // display the 'scratchpad' variable fields

      // test if temperature is .0 or .5
      decimal = scratchpad[0] & (1 << 0);  // decimals is in bit 0, store value
      // move temperature bits on position to the right
      scratchpad[0]>>=1;

      Write_Serial_Number(scratchpad[0]); 
      Write_Serial_Data(46);                    // '.'

      if (decimal == 0)
      {
        Write_Serial_Data(48);   // .0
      } else {
        Write_Serial_Data(53);   // .5
      }

      Write_Serial_Data(32);                    // ' '
      Write_Serial_Data(67);                    // 'C'

   
      _delay_ms(1000);

      device=device+1;

    } else {

      break;

    }

  } 

}

// -- end OneWire functions --



int main()
{
  initialise_Serial(9600);

  _delay_ms(1000);
  _delay_ms(1000);

  char Message1[] = "The quick brown fox jumps over the lazy dog\r\n";

  while (1)
  {   

    Write_Serial_Text(Message1);
    _delay_ms(1000);
    _delay_ms(1000);
    _delay_ms(1000);

  }
 
  return 0;
}

int main2()
{

  uint8_t  ack;
  int devs;

  initialise_Serial(9600);

  _delay_ms(1000);
  _delay_ms(1000);



  char Message1[] = "The quick brown fox jumps over the lazy dog\r\n";

  while (1)
  {   

    Write_Serial_Text(Message1);
    _delay_ms(1000);
    _delay_ms(1000);
    _delay_ms(1000);

  }
 

  while (1)
  {

    ack = owReset(THERM_DQ);

    if (ack == 1)
    {

      strcpy_P(buffer, (char*)pgm_read_word(&(msg[01]))); 
      Write_Serial_Text(buffer);  // 'No sensors found'
      _delay_ms(1000);

    } else {

      // sensor activity detected 

      while (1)
      {

        devs=scanbus();  // detect and store device rom's in an array

        if (devs > 0)
        {

          strcpy_P(buffer, (char*)pgm_read_word(&(msg[04]))); 

          Write_Serial_Text(buffer);  // 'Devices: '
          Write_Serial_Number(devs);     
          _delay_ms(1000);
          _delay_ms(1000);
 
          showdevs();

        }

        _delay_ms(1000);
        _delay_ms(1000);

      }

    }

  }

  /*

           DS18S20

           byte 1                 byte 2
           -------------------------------
           76543210             76543210
           00101010             00000000
           IIIIDDDD             SSSSSIII 
           3210                      654 

           I = Celsius integer digits (signed)          000 0010 = 2
           D = Number of decimal steps (x multiplier)   1010     = 10 steps x 0.0625
           S = Signedness (0=positive, 1=negative)      00000    = +
   

           DS1820

           LSB (1st 7 bits are the temperature)
           and the last bit represents .0 or .5 degrees

           Example:   '87654321'
                       TTTTTTTH
           
             T = Temperature
             H = halves H=0 (.0), H=1 (.5)

           MSB (is 0 when positive, 1 when negative) 

  */

  return 0;

}
